package com.abssqr.plat.core.service.plan;

import com.abssqr.plat.common.facade.enums.*;
import com.abssqr.plat.common.facade.model.plan.*;
import com.abssqr.plat.common.model.enums.IdTypeEnum;
import com.abssqr.plat.common.model.repo.plan.PlanAcctTitleRepo;
import com.abssqr.plat.common.model.repo.plan.PlanAllocRepo;
import com.abssqr.plat.common.model.seq.SequenceUtil;
import com.general.system.common.util.DateTimeUtil;
import com.general.system.common.util.VarChecker;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.stream.Collectors;

/**
 * 生成对应的兑付计划
 */
@Component
public class PlanAllocManager {


    @Autowired
    PlanAllocRepo planAllocRepo;

    @Autowired
    PlanAcctTitleRepo planAcctTitleRepo;

    /**
     * 初始化兑付计划
     * @param plan
     * @param status
     */
    public void initPlanAlloc(Plan plan, PlanStatusEnum status) {
        if (status == PlanStatusEnum.PREPARING) {
            planAllocRepo.cleanByPlanNo(plan.getPlanNo());
        }
        List<PlanAlloc> planAllocList = caclPlanAllocByDate(plan);
        //根据分配日进行排序
        List<PlanAlloc> sortedPlanAlllocList = planAllocList.stream()
                .sorted(Comparator.comparing(PlanAlloc::getAllocDate))
                .collect(Collectors.toList());

        Map<String, PlanAcctTitle> planAcctTitlerMap = getPlanAcctTitleMap(plan.getPlanNo());

        //根据投资结构生成兑付计划
        calcInvestStructureAlloc(plan.getInvestStructure(), sortedPlanAlllocList,planAcctTitlerMap);

        //根据费用结构生成兑付计划
        caclFeeStructureAlloc(plan.getFeeStructure(), sortedPlanAlllocList,planAcctTitlerMap);

        planAllocRepo.store(sortedPlanAlllocList);
    }

    /**
     * 根据首次计算日生成分配计划
     *
     * @param plan
     * @return
     */
    private List<PlanAlloc> caclPlanAllocByDate(Plan plan) {

        List<PlanAlloc> result = new ArrayList<>(5);
        //首次兑付日
        Date firstAllocDate = plan.getFirstAllocDate();
        //首次计算日
        Date firstCalcDate = plan.getFirstCalcDate();
        //首次转付日
        Date firstPayDate = plan.getFirstPayDate();
        //首次报告日
        Date firstReportDate = plan.getFirstReportDate();
        //首次结息日
        Date firstSettleDate = plan.getFirstSettleDate();

        Date planEndDate = plan.getEndDate();
        //下次计算日
        Date nextCaclDate = (Date) firstCalcDate.clone();
        //下次兑付日
        Date nextAllocDate = (Date) firstAllocDate.clone();
        //下次转付日
        Date nextPayDate = (Date) firstPayDate.clone();
        //下次报告日
        Date nextReportDate = (Date) firstReportDate.clone();
        //下次结息日
        Date nextSettleDate = (Date) firstSettleDate.clone();

        int interval = plan.getPayFreq().getCnt();

        //兑付日小于产品到期日
        while (DateTimeUtil.isBeforeOnDate(DateTimeUtil.addMonths(nextAllocDate, interval),planEndDate)){
            PlanAlloc planAlloc = buildBasicPlanAlloc(plan.getPlanNo());

            nextCaclDate = DateTimeUtil.addMonths(nextCaclDate, interval);
            planAlloc.setCalcDate(nextCaclDate);

            nextAllocDate=DateTimeUtil.addMonths(nextAllocDate, interval);
            planAlloc.setAllocDate(nextAllocDate);

            nextPayDate = DateTimeUtil.addMonths(nextPayDate, interval);
            planAlloc.setPayDate(nextPayDate);

            nextSettleDate= DateTimeUtil.addMonths(nextSettleDate, interval);
            planAlloc.setSettleDate(nextSettleDate);

            nextReportDate=DateTimeUtil.addMonths(nextReportDate, interval);
            planAlloc.setReportDate(nextReportDate);

            result.add(planAlloc);
        }

        PlanAlloc planAlloc = buildBasicPlanAlloc(plan.getPlanNo());

        nextCaclDate = DateTimeUtil.addMonths(nextCaclDate, interval);

        planAlloc.setCalcDate(nextCaclDate);
        nextAllocDate=DateTimeUtil.addMonths(nextAllocDate, interval);
        planAlloc.setAllocDate(nextAllocDate);

        nextPayDate = DateTimeUtil.addMonths(nextPayDate, interval);
        planAlloc.setPayDate(nextPayDate);

        nextSettleDate= DateTimeUtil.addMonths(nextSettleDate, interval);
        planAlloc.setSettleDate(nextSettleDate);

        nextReportDate=DateTimeUtil.addMonths(nextReportDate, interval);
        planAlloc.setReportDate(nextReportDate);
        result.add(planAlloc);

        return result;
    }


    /**
     * 根据投资结构生成对应的本金兑付事项和利息兑付事项
     *
     * @param investStructureList
     * @param allocList
     */
    private void calcInvestStructureAlloc(List<InvestStructure> investStructureList, List<PlanAlloc> allocList, Map<String,PlanAcctTitle> acctTitleMap) {

        for (InvestStructure investStructure : investStructureList) {
            AllocMethodEnum allocMethod = investStructure.getAllocMethod();
            Date effectDate = investStructure.getEffectDate();
            Date endDate = investStructure.getEndDate();
            for (PlanAlloc planAlloc : allocList) {
                //判断分配日是否大于投资结构起息日
                if (DateTimeUtil.between(planAlloc.getAllocDate(),effectDate , endDate, DateTimeUtil.BetweenType.se)) {
                    //如果是过手型生成还本科目
                    if (allocMethod == AllocMethodEnum.PT) {
                        TitleTypeEnum titleType=investStructure.getLayerType()== LayerTypeEnum.PRIOR?
                                TitleTypeEnum.PRIOR_INVESTED:
                                TitleTypeEnum.SECND_INVESTED;
                        //生成本金科目
                        appendPlanAllocItem(planAlloc,investStructure.getLayerNo(),titleType,acctTitleMap);
                    }


                }
                /**
                 * 如果是先息后本型 则在计划结束日前一期的分配日
                 */
                if(DateTimeUtil.isBeforeOnDate(endDate,planAlloc.getAllocDate())){
                    if(allocMethod == AllocMethodEnum.IBC){
                        TitleTypeEnum titleType=investStructure.getLayerType()== LayerTypeEnum.PRIOR?
                                TitleTypeEnum.PRIOR_INVESTED:
                                TitleTypeEnum.SECND_INVESTED;

                        //生成本金科目
                        appendPlanAllocItem(planAlloc,investStructure.getLayerNo(),titleType,acctTitleMap);
                        break;
                    }

                }

                //生成对应的利息兑付事项
                if(DateTimeUtil.between(planAlloc.getAllocDate(), effectDate, endDate, DateTimeUtil.BetweenType.se)){
                    appendPlanAllocItem(planAlloc,investStructure.getLayerNo(),TitleTypeEnum.INVEST_PROFIT,acctTitleMap);
                }

            }
        }
    }

    /**
     * 根据费用结构生成兑付计划
     * @param feeStructureList
     * @param allocList
     * @param acctTitleMap
     */
    private void caclFeeStructureAlloc(List<FeeStructure> feeStructureList, List<PlanAlloc> allocList, Map<String,PlanAcctTitle> acctTitleMap){


        for(FeeStructure feeStructure:feeStructureList){
            FeeTimeTypeEnum feeCalcType = feeStructure.getFeeTimeType();
            switch (feeCalcType){
                //项目到期日
                case PEDD:{
                    //获取最后一期的兑付计划
                    PlanAlloc planAlloc = allocList.get(allocList.size() - 1);
                    appendPlanAllocItem(planAlloc, feeStructure.getFeeNo(), TitleTypeEnum.FEE, acctTitleMap);
                    break;
                }
                //首次兑付日(首次分配日)
                case PFCD:{
                    //获取第一期的兑付计划
                    PlanAlloc planAlloc = allocList.get(0);
                    appendPlanAllocItem(planAlloc, feeStructure.getFeeNo(), TitleTypeEnum.FEE, acctTitleMap);
                    break;
                }
                //按期分配
                case TERM:{
                    allocList.forEach(planAlloc -> {
                        appendPlanAllocItem(planAlloc, feeStructure.getFeeNo(), TitleTypeEnum.FEE, acctTitleMap);
                    });
                    break;
                }
            }
        }
    }


    private void appendPlanAllocItem(PlanAlloc planAlloc, String refNo,
                                             TitleTypeEnum titleType, Map<String, PlanAcctTitle> planAcctTitleMap){
        PlanAllocItem planAllocItem = buildPlanAllocItem(planAlloc.getPlanNo(), planAlloc.getAllocNo(), refNo, titleType, planAcctTitleMap);
        planAlloc.getPlanAllocItems().add(planAllocItem);
    }

    private Map<String, PlanAcctTitle> getPlanAcctTitleMap(String planNo) {
        List<PlanAcctTitle> planAcctTitles = planAcctTitleRepo.getByEntityNo(planNo);
        VarChecker.checkNotNull(planAcctTitles, "计划[{0}]不存在对应的科目信息", planNo);
        Map<String, PlanAcctTitle> planAcctTitleMap = planAcctTitles.stream().collect(Collectors.toMap((planAcctTitle) -> {
            return planAcctTitle.getRefNo() + planAcctTitle.getTitleType().getCode();
        }, (planAcctTitle) -> {
            return planAcctTitle;
        }));
        return planAcctTitleMap;
    }





    private PlanAlloc buildBasicPlanAlloc(String planNo) {
        PlanAlloc planAlloc = new PlanAlloc();
        planAlloc.setPlanNo(planNo);
        planAlloc.setAllocNo(SequenceUtil.genId(IdTypeEnum.ALLOCPRG));

        return planAlloc;
    }





    private PlanAllocItem buildPlanAllocItem(String planNo, String allocNo, String refNo,
                                             TitleTypeEnum titleType, Map<String, PlanAcctTitle> planAcctTitleMap) {
        AllocTypeEmun allocType = titleType.getAllocType();
        PlanAcctTitle planAcctTitle = planAcctTitleMap.get(refNo + titleType.getCode());
        VarChecker.checkNotNull(planAcctTitle, "[{0}]-[{2}]-[{1}]没有找到对应科目", planNo, refNo,
                titleType.getCode());

        PlanAllocItem planAllocItem = new PlanAllocItem();
        planAllocItem.setPlanNo(planNo);
        planAllocItem.setItemNo(SequenceUtil.genId(IdTypeEnum.ALLOCITEM));
        planAllocItem.setAllocNo(allocNo);
        planAllocItem.setSort(allocType.getSort());
        planAllocItem.setAllocType(allocType);
        planAllocItem.setTitleName(planAcctTitle.getTitleName());
        planAllocItem.setTitleNo(planAcctTitle.getTitleNo());
        return planAllocItem;
    }

}
